/******************************************************************************/
/*   IBM Confidential                                                         */
/*                                                                            */
/*   OCO Source Materials.                                                    */
/*                                                                            */
/*   (C) Copyright IBM Corp. 2001.                                            */
/*                                                                            */
/*   The source code for this program is not published or otherwise           */
/*   divested of its trade secrets, irrespective of what has been             */
/*   deposited with the U.S. Copyright Office.                                */
/*                                                                            */
/*   Product ID:                                                              */
/******************************************************************************/

/**
 * Descriptive Name:  NursesAid.java
 *
 * Description:
 *
 *
 *    Modifications:
 *
 *    Date      Author                 Category             Defect/Feature
 *              Description
 *
 *  08/08/2001  Dave Been              Creation
 *
 */

package com.ibm.mobileservices.demo;

import com.ibm.mobileservices.lang.SampleCLPmsg;

/**
 * CS version of DB2 Everyplace Virtual Nurse
 *
 *		This class is the MIDlet run inside the device
 */

import javax.microedition.midlet.*;
import javax.microedition.lcdui.*;
import javax.microedition.rms.*;
import javax.microedition.io.Connector;
import javax.microedition.io.HttpConnection;
import java.lang.Exception;
import java.io.IOException;
import java.io.EOFException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.DataOutputStream;
import java.io.DataInputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.util.Hashtable;
import java.util.Vector;

import com.ibm.mobileservices.isync.ISync;
import com.ibm.mobileservices.isync.ISyncDriver;
import com.ibm.mobileservices.isync.midp.MIDPISyncProvider;
import com.ibm.mobileservices.isync.ISyncException;
import com.ibm.mobileservices.isync.event.ISyncEvent;
import com.ibm.mobileservices.isync.event.ISyncListener;
import com.ibm.mobileservices.isync.ISyncSubscriptionSet;
import com.ibm.mobileservices.isync.midp.FastRecordStore;
import com.ibm.mobileservices.isync.midp.FastRecordEnumeration;
import com.ibm.mobileservices.isync.midp.Index;

import com.ibm.mobileservices.isync.midp.TableMetaData;

import com.ibm.mobileservices.isync.debug.Debug;
public class NursesAid
{
	public static final int VNSCHEDULE 		= 1;
	public static final int VNPERSON 			= 2;
	public static final int VNMEDICALRECORD	= 3;
	public static final int VNCONTACT			= 4;

	private Vector schedIds;		// key for future lookups in RMS
	String [] schedule;				// schedule row for later name lookup.

	private String patientId;
	private String patientName;

	private Hashtable persons;		// hash id->name
	private Hashtable schedules;	// hash timeStr->id
	private boolean justPersonName;	// only read id/name from VNPERSON

	/*
		Share AddRecord/UpdateRecord code, does addRecord/setRecord
		in readMedRecord puts the row into a TextField
	*/
	private boolean updateRecord;

	private boolean medVerbose;		// display whole record if verbose, not synopsis
	private int maxRecId;			// used if user inserts a new record

	// used by readRecords
	private Vector aVector;

	private Vector contactIds;
	private int contactChoosen;
	private int schedChoosen;
	private int recChoosen;			// index in choice as to selected record
	private Vector recordIds;		// recordIds to choose/lookup in View button
	private int recordId;			// picked recid to view, update, delete

	ChoiceGroup schedChoice;
	ChoiceGroup recChoice;
	ChoiceGroup contactChoice;

	private Display display;		// The display for this MIDlet

	private Form currentForm;
	private Form statusForm;

	private TextField bloodPressTF;
	private TextField pulseRateTF;
	private TextField tempTF;
	private TextField weightTF;
	private TextField commentTF;

	private ISyncDriver isync;

	private FastRecordStore rms;

	// the goal is to alloc as few objects as possible, see RMS usage
	private byte data[] = new byte[80];

	private String[] strings;

	ByteArrayInputStream bin;
	DataInputStream din;

	/**
		Per Form/Store methods
	 */
	public NursesAid(ISyncDriver isync,
		Display display, Form statusForm)
	{
		this.isync = isync;
		this.display = display;
		this.statusForm = statusForm;
	}
	public void setISync(ISyncDriver isync)
	{
		this.isync = isync;
	}
	Form showSchedule(Command[] cmds)
		throws Exception
	{
		Form schedForm = null;

		if (Debug.DEBUG)
			System.out.println("Inside showSchedule, cmds " + cmds);
		patientId = null;
		schedChoosen = 0;

		/**
			if going back to this form w/o data change
			just use existing schedIds
			In the future if rows change, have to set schedIds to null.
		*/
		if (schedIds == null)
		{
			schedIds = new Vector();
			persons = new Hashtable();
			schedules = new Hashtable();

			try
			{
				/**
					We have to join PERSON and SCHEDULE by patient id
					in order to get the name for the schedule.
					Ideally we could do the join in the server, but ....
				*/
				justPersonName = true;
				readRecords(SampleCLPmsg.VNPERSON, VNPERSON, null);
				justPersonName = false;

				// if (Debug.DEBUG && Debug.LEVEL > 5)
					// System.out.println("\t\tpersons prop: " + persons);

				// read PERSON 1st so doesn't zap aVector
				schedule = readRecords(SampleCLPmsg.VNSCHEDULE, VNSCHEDULE, null);
			}
			catch (Exception e)
			{
				schedIds = null;
				throw e;
			}
			if (schedule == null)
			{
				// reset to read next time
				schedIds = null;
				return null;
			}

			// aVector is vector of times
			// schedules is hash table of times->ids
			// persons is hash table of id->names

			if (aVector != null)
			{
				int avecSize = aVector.size();
				schedule = new String[avecSize];
				// System.out.println(">>>>  convert aVector to schedule array "
					// + aVector.size());

				// do a JOIN, sigh, wish there was a database in the house
				for (int i = 0; i < avecSize; i++)
				{
					String tm = (String) aVector.elementAt(i);
					// System.out.println("tm " + tm);
					String id = (String) schedules.get(tm);
					// System.out.println("id " + id);
					String name = (String) persons.get(id);
					// System.out.println("name " + name);
					schedule[i] = tm + " " + name;
				}
			}
		}

		schedChoice = showChoiceGroup(schedule, schedForm, SampleCLPmsg.YourSchedule, SampleCLPmsg.Today, cmds);
		if (Debug.DEBUG)
			System.out.println(" sched choice is " + schedChoice);
		schedForm = currentForm;
		return schedForm;
	}

	Form showInfo(Form infoForm, Command[] cmds)
	{
		try
		{
			if (Debug.DEBUG)
				System.out.println(" showInfo choose " + schedChoosen);
			patientId = (String) schedIds.elementAt(schedChoosen);
			String sched = schedule[schedChoosen];
			int index = sched.indexOf(" ");
			// int endIndex = sched.indexOf(" ", index+1);
			patientName = sched.substring(index + 1); // , endIndex);
			if (Debug.DEBUG)
				System.out.println(" showInfo patientId " + patientId + " " + patientName);

			// RecordFilter rf = new PersonRecordFilter(patientId);
			String [] people = findRecords(SampleCLPmsg.VNPERSON, VNPERSON);
			infoForm = showForm(people, infoForm, SampleCLPmsg.PersonInfo, 
				SampleCLPmsg.Info, cmds);
		}
		catch (Exception e)
		{
			if (Debug.DEBUG)
				System.out.println("showInfo exc: " + e.toString());
			if (Debug.DEBUG && Debug.LEVEL > 8)
				e.printStackTrace();
		}
		return infoForm;
	}
	Form showContactInfo(Form contactInfoForm, Command[] cmds)
	{
	// System.out.println(" in showContactInfo");
		try
		{
			String contactId = (String) contactIds.elementAt(contactChoosen);
			if (Debug.DEBUG)
				System.out.println(" showContactInfo contactId " + contactId);

			String savePatientId = patientId;
			patientId = contactId;
			String [] people = findRecords(SampleCLPmsg.VNPERSON, VNPERSON);
			patientId = savePatientId;

			contactInfoForm = showForm(people, contactInfoForm,
					SampleCLPmsg.PersonInfo, SampleCLPmsg.Info, cmds);
		}
		catch (Exception e)
		{
			if (Debug.DEBUG)
				System.out.println("showContactInfo exc: " + e.toString());
			if (Debug.DEBUG && Debug.LEVEL > 8)
				e.printStackTrace();
		}
		return contactInfoForm;
	}
	Form showRecords(Form recForm, Command[] cmds)
	{
		if (Debug.DEBUG)
			System.out.println(" >>>>>> showRecords");
		recordId = -1;				// if user doesn't change selection default to first record...?
		recChoosen = 0;

		if (recordIds == null)
			recordIds = new Vector();
		else
			recordIds.removeAllElements();

		medVerbose = false;
		maxRecId = -1;
		String [] medRecs = findRecords(SampleCLPmsg.VNMEDICALRECORD, VNMEDICALRECORD);
		try
		{
			recChoice = showChoiceGroup(medRecs, recForm,
				SampleCLPmsg.MedRecList, patientName, cmds);
			recForm = currentForm;
		}
		catch (Exception e)
		{
			if (Debug.DEBUG)
				System.out.println("show med record list failed " 
					+ e.toString());
			if (Debug.DEBUG && Debug.LEVEL > 8)
				e.printStackTrace();
			return null;
		}
		// System.out.println(" >>>>>>>>>>>>>>>> showRecords returns recForm "
			// + recForm);
		return recForm;
	}
	
	public void setDirty(String storeName)
	{
		FastRecordStore rms = null;
		Index index = null;

		if (isync == null)
		{
			System.out.println("rmsOperation isync is null");
			return;
		}
		try
		{
			rms = FastRecordStore.openRecordStore(storeName, false);
			TableMetaData rs = ((MIDPISyncProvider)isync).getTableMetaDataByName(storeName);
			index = new Index(rms, rs);

			FastRecordEnumeration enum =
							rms.enumerateRecords(null, null, false);
			int recSize;

			while (enum.hasNextElement())
			{
				System.out.println(" setDiry get next id");
				int id = enum.nextRecordId();

				recSize = rms.getRecordSize(id);
				checkDataSizeOpenStream(recSize);
				recSize = rms.getRecord(id, data, 0);
				System.out.println("set DELETED DIRTY");
				data[0] = ISync.ROW_DELETED;
				index.updateRecord(id, data, recSize);
			}
			// index.close();
			index = null;
			rms.closeRecordStore();
			rms = null;
		}
		catch (Exception e)
		{
			System.out.println("	setDirty failed " + e.toString());
		}
		finally
		{
			try
			{
				if (rms != null)
					rms.closeRecordStore();
				if (index != null)
					index.close();
				rms = null;
			}
			catch (Exception e) { }
		}
	}

	private static final byte OP_READ		= 1;
	private static final byte OP_DELETE		= 2;

	private Object rmsOperation(int recId, byte operation)
	{
		/*
			We know the selected 'recordId' in RMS already, just get it
		*/

		FastRecordStore rms = null;
		Index index = null;

		if (isync == null)
		{
			System.out.println("rmsOperation isync is null");
			return null;
		}
		try
		{
			rms = FastRecordStore.openRecordStore(SampleCLPmsg.VNMEDICALRECORD, false);
			TableMetaData rs = ((MIDPISyncProvider)isync).getTableMetaDataByName(SampleCLPmsg.VNMEDICALRECORD);


			// System.out.println("rmsop: got tmd " + rs);
			index = new Index(rms, rs);
			int recSize;
			switch (operation)
			{
			  // used by modifyRecord (add/update) and showRecDetail()
			  case OP_READ:
			/* Read only operations here, just use the rms directly */

				recSize = rms.getRecordSize(recordId);
				checkDataSizeOpenStream(recSize);
				rms.getRecord(recordId, data, 0);
				rms.closeRecordStore();
				rms = null;

				// discard 1-byte row marker
				din.readByte();

				medVerbose = true;
				String s = readMedRecord(recordId, din);
				medVerbose = false;
				return s;

			  case OP_DELETE:
			 /* We set the dirty bit to true to indicate a "soft" delete */
			 System.out.println(" OP_DELETE");
				recSize = rms.getRecordSize(recordId);
				checkDataSizeOpenStream(recSize);
				recSize = rms.getRecord(recordId, data, 0);
				System.out.println("set DELETED DIRTY");
				data[0] = ISync.ROW_DELETED;
				index.updateRecord(recordId, data, recSize);
				index.close();
				index = null;
				rms.closeRecordStore();
				rms = null;
				break;
			}
		}
		catch (Exception e)
		{
			currentForm.append(e.toString());
			if (Debug.DEBUG)
				System.out.println("rmsOperation: findRecord failed " + e.toString());
			if (Debug.DEBUG && Debug.LEVEL > 8)
				e.printStackTrace();
		}
		finally
		{
			try
			{
				if (rms != null)
					rms.closeRecordStore();
				if (index != null)
					index.close();
				rms = null;
			}
			catch (Exception e) { }
		}
		return null;
	}
	private String[] sArray;

	Form showRecDetails(Form recDetailForm, Command[] cmds)
	{
		if (Debug.DEBUG)
			System.out.println(" showRecDetails " + recordId);

		try
		{
			String s = (String) rmsOperation(recordId, OP_READ);
			if (s == null)
				s = "No record from rmsOp";
			System.out.println(" read " + s);

			if (sArray == null)
				sArray = new String[1];
			sArray[0] = s;

			recDetailForm = showForm(sArray, recDetailForm,
				SampleCLPmsg.MedicalRecord, patientName, cmds);
		}
		catch (Exception e)
		{
			medVerbose = false;
			if (Debug.DEBUG)
				System.out.println("show recDetail failed " + e.toString());
			if (Debug.DEBUG && Debug.LEVEL > 8)
				e.printStackTrace();
		}
		if (Debug.DEBUG)
			System.out.println(" showRecDetails EXIT " + recordId);
		return recDetailForm;
	}
	
	Form modifyRecord(Form addForm, Command saveButton, Command backButton,
		boolean update, boolean performingUpdate)
	{
		if (Debug.DEBUG)
			System.out.println("modifyRecord update: " + update
				+ " performingUpdate " + performingUpdate);

		/*
			This method sets up the add/change form, we keep track
			of this so when we read the medicalrecords, we put the row
			(based on recordId), into the TextField
		*/

		updateRecord = performingUpdate;

		if (addForm == null)
			addForm = new Form(SampleCLPmsg.NewMedicalRecord);
		else
			clearDisplay(addForm);

		addForm.append(patientName);

		if (!update)
			setTextFields("", "", "", "", "");
		else
		{
			// read the fields into the textField entries
			rmsOperation(recordId, OP_READ);
		}

		addForm.append(bloodPressTF);
		addForm.append(pulseRateTF);
		addForm.append(tempTF);
		addForm.append(weightTF);
		addForm.append(commentTF);

		addForm.addCommand(saveButton);
		addForm.addCommand(backButton);

		display.setCurrent(addForm);
		return addForm;
	}
	private void setTextFields(String bp, String pr, String temp,
		String w, String com)
	{
		if (bloodPressTF == null)
			bloodPressTF = new TextField(SampleCLPmsg.BloodPressure, bp, 6, TextField.ANY);
		else
			bloodPressTF.setString(bp);
		if (pulseRateTF == null)
			pulseRateTF = new TextField(SampleCLPmsg.PulseRate, pr,
				8, TextField.NUMERIC);
		else
			pulseRateTF.setString(pr);

		// ANY since '.' is not NUMERIC
		if (tempTF == null)
			tempTF = new TextField(SampleCLPmsg.Temperature, temp,
				5, TextField.ANY);
		else
			tempTF.setString(temp);

		// ANY since '.' is not NUMERIC, 170.00 is one example, too big
		if (weightTF == null)
			weightTF = new TextField(SampleCLPmsg.Weight, w, 7, TextField.ANY);
		else
			weightTF.setString(w);
		if (commentTF == null)
			commentTF = new TextField(SampleCLPmsg.Comment, com, 100, TextField.ANY);
		else
			commentTF.setString(com);
	}

	void saveRecord()
	{
		if (Debug.DEBUG)
			System.out.println(" saveRecord");
		String bloodpres 	= bloodPressTF.getString();
		String pulseRate 	= pulseRateTF.getString();
		String temp 		= tempTF.getString();
		String weight 		= weightTF.getString();
		String comment 		= commentTF.getString();

		if (bloodpres == null || bloodpres.length() == 0)
			bloodpres = "99/99";
		if (pulseRate == null || pulseRate.length() == 0)
			pulseRate = "99";
		if (temp == null || temp.length() == 0)
			temp = "99";
		if (weight == null || weight.length() == 0)
			weight = "99";
		if (comment == null || comment.length() == 0)
			comment = SampleCLPmsg.nocomment;

		if (Debug.DEBUG)
		{
			System.out.println(" bloodpres " + bloodpres);
			System.out.println(" pulse " + pulseRate);
			System.out.println(" temp " + temp);
			System.out.println(" weight " + weight);
			System.out.println(" comment " + comment);
			System.out.println(" patientId " + patientId);
		}

		ByteArrayOutputStream bout = null;
		DataOutputStream dout = null;
		clearDisplay();
		try
		{
			bout = new ByteArrayOutputStream();
			dout = new DataOutputStream(bout);

			// This indicates the target created this row
			if (updateRecord)
			{
				if (Debug.DEBUG && Debug.LEVEL > 5)
					System.out.println(" write CHANGED DIRTY");
				dout.writeByte(ISync.ROW_CHANGED);
				dout.writeInt(recordId);
			}
			else
			{
				dout.writeByte(ISync.ROW_ADDED);
				maxRecId += 1;
				if (Debug.DEBUG && Debug.LEVEL > 5)
					System.out.println("  NEW RECORD " + maxRecId 
						+ "  patid " + patientId);

				dout.writeInt(maxRecId);
			}

			dout.writeBoolean(false);
			// "2002-3-31"
			dout.writeInt((2002 << 16) + (3 << 8) + 31);

			dout.writeBoolean(false);

			// "12:00:00"
			dout.writeInt((12 << 16));

			// dout.writeBoolean(false);
			dout.writeUTF(patientId);

			dout.writeBoolean(false);
			dout.writeUTF(bloodpres);

			short pulse = -1;
			dout.writeBoolean(false);
				// dout.writeUTF(pulseRate);
			try
			{
				pulse = Short.parseShort(pulseRate);
			}
			catch (Exception se)
			{
				System.out.println("Error parsing short " + se.toString());
			}
			dout.writeShort(pulse);

			dout.writeBoolean(false);
			dout.writeUTF(temp);

			dout.writeBoolean(false);
			dout.writeUTF(weight);

			dout.writeBoolean(false);
			dout.writeUTF(comment);
		}
		catch (Exception e)
		{
			currentForm.append("Failed to Write Record to stream " + e.toString());
			try
			{
				Thread.sleep(2000);
			} catch (/* Interrupted */ Exception ee) { }
			if (Debug.DEBUG)
				System.out.println("Failed to add record " 
					+ e.toString() + " " + e.getMessage());
			if (Debug.DEBUG && Debug.LEVEL > 8)
				e.printStackTrace();
		}
		Index index = null;
		try
		{
			rms = FastRecordStore.openRecordStore(SampleCLPmsg.VNMEDICALRECORD, false);
			index = new 
				Index(rms,((MIDPISyncProvider)isync).getTableMetaDataByName(SampleCLPmsg.VNMEDICALRECORD));
			if (Debug.DEBUG)
				System.out.println("DEBUG: Store size avail " 
					+ rms.getSizeAvailable());
			int len = bout.size();
			if (Debug.DEBUG)
				System.out.println(" Write # bytes " + len);

			String msg;
			if (updateRecord)
			{
				if (Debug.DEBUG && Debug.LEVEL > 8)
					System.out.println("doing an update of new record...");
				index.updateRecord(recordId, bout.toByteArray(), len);
				msg = "Updated Record";
			}
			else
			{
				if (Debug.DEBUG && Debug.LEVEL > 8)
					System.out.println("doing an insert of new record...");
				index.insertRecord(bout.toByteArray(), len);
				msg = "Added Record";
			}
			currentForm.append(msg);

			try
			{
				Thread.sleep(2000);
			} catch (/* Interrupted */ Exception ee) { }
		}
		catch (Exception e)
		{
			currentForm.append("Failed to Add/Update Record " + e.toString());
			try
			{
				Thread.sleep(2000);
			} catch (/* Interrupted */ Exception ee) { }
			if (Debug.DEBUG)
				System.out.println("Failed to add record "
					+ e.toString() + " " + e.getMessage());

					
			if (Debug.DEBUG && Debug.LEVEL > 8)
				e.printStackTrace();
		}
		finally
		{
			try
			{
				if (dout != null)
					((OutputStream) dout).close();
				if (bout != null)
					bout.close();
				if (rms != null)
					rms.closeRecordStore();
				if (index != null)
					index.close();
			}
			catch (Exception ee) { }
			updateRecord = false;
		}
	}
	
	void deleteRecord()
	{
		if (Debug.DEBUG)
			System.out.println("deleteRecord " + recordId);
		clearDisplay();

		rmsOperation(recordId, OP_DELETE);
	}
	Form showContacts(Form contactForm, Command[] cmds)
	{
		if (Debug.DEBUG)
			System.out.println(" showContacts");

		contactChoosen = 0;

		if (contactIds == null)
			contactIds = new Vector();
		else
			contactIds.removeAllElements();

		String [] contactsRecs = findRecords(SampleCLPmsg.VNCONTACT, VNCONTACT);
		try
		{
			contactChoice = showChoiceGroup(contactsRecs, contactForm,
				SampleCLPmsg.EmergContacts, patientName, cmds);
			contactForm = currentForm;

		}
		catch (Exception e)
		{
			if (Debug.DEBUG)
				System.out.println("showContacts failed " + e.toString());
			if (Debug.DEBUG && Debug.LEVEL > 8)
				e.printStackTrace();
			return null;
		}
		return contactForm;
	}

	/**
		Table specific RMS readers
	 */
	private String readSchedule(DataInputStream din) throws IOException
	{
		if (Debug.DEBUG && Debug.LEVEL > 10)
			System.out.println("\treadSchedule avail " + din.available());

		// As of 7/12/2002 no null on patientid
		// din.readBoolean();
		String patientId = din.readUTF();
		// if (Debug.DEBUG && Debug.LEVEL > 10)
			// System.out.println("\t patientId " + patientId);

		// primary key not null, no din.readBoolean();
		// Date time_c = new java.util.Date(din.readLong());
		int time_c = din.readInt();
		// if (Debug.DEBUG && Debug.LEVEL > 10)
			// System.out.println("time_c is  " + time_c);

		String timeStr = fmtTime(time_c);
		// if (Debug.DEBUG && Debug.LEVEL > 10)
			// System.out.println("\t timeStr " + timeStr);

		schedIds.addElement(patientId);
		schedules.put(timeStr, patientId);

		return timeStr;
	}
	private String readPerson(DataInputStream din) throws IOException
	{
	// System.out.println("\t\treadPerson called");

		// not null primary key: din.readBoolean();
		String id = din.readUTF();

		din.readBoolean();
		String name = din.readUTF();
		if (Debug.DEBUG && Debug.LEVEL > 5)
			System.out.println("\t\treadPerson: " + id + " name " + name);
		if (justPersonName)
		{
			persons.put(id, name);
			return null;
		}
		din.readBoolean();
		String addr = din.readUTF();
		din.readBoolean();
		String city = din.readUTF();
		din.readBoolean();
		String hmPhone = din.readUTF();
		din.readBoolean();
		String wkPhone = din.readUTF();
		din.readBoolean();
		String mbPhone = din.readUTF();
		din.readBoolean();

		StringBuffer buf = new StringBuffer(SampleCLPmsg.Name);
		buf.append(name).
			append(SampleCLPmsg.Address).append(addr).
			append(SampleCLPmsg.City).append(city).
			append(SampleCLPmsg.Home).append(hmPhone).
			append(SampleCLPmsg.Work).append(wkPhone).
			append(SampleCLPmsg.Mobile).append(mbPhone);

		String ret = buf.toString();

		if (Debug.DEBUG && Debug.LEVEL > 5)
			System.out.println(ret);
		return ret;
	}
	private String readMedRecord(int rmsRecId, DataInputStream din)
		throws IOException
	{
		// 		Table columns
		// RecordID Integer not null primary key,
		// Date_C date,                              
		// Time_C datetime hour to second ,  
		// PatientID Varchar(9) not null,            
		// BloodPressure Char(7) ,           
		// PulseRate Smallint ,              
		// Temperature Char(5) ,             
		// Weight Char(7) ,
		// Comment Varchar(100),            

		int recId = din.readInt();

		// track max for possible AddRecord by user
		if (recId > maxRecId)
			maxRecId = recId;
		
		// set up the first selection default recordId to the rms Id.
		if (recordId == -1)
			recordId = rmsRecId;

		recordIds.addElement(new Integer(rmsRecId));

		din.readBoolean();
		int date_c = din.readInt();
		String dateStr = fmtDate(date_c);

		din.readBoolean();
		int time_c = din.readInt();
		String timeStr = fmtTime(time_c);

		StringBuffer buffy = null;
		
		if (medVerbose)
		{
			// As of 7/12/2002 no null on patientid
			// din.readBoolean();

			// read PatientId
			din.readUTF();

			din.readBoolean();
			String bloodPress = din.readUTF();
			din.readBoolean();
			short pulseRate = din.readShort();
			din.readBoolean();
			String temp = din.readUTF();
			din.readBoolean();
			String weight = din.readUTF();
			din.readBoolean();
			String comment = din.readUTF();

			// Just put into text fields of Form
			if (updateRecord)
			{
				String pulse = "-1";

				pulse = String.valueOf(pulseRate);

				setTextFields(bloodPress, pulse, temp, weight, comment);
				return null;
			}
			buffy = new StringBuffer(patientName);

			buffy.append("\n").append(dateStr).append(" ").append(timeStr)
				.append(SampleCLPmsg.BloodPrs).append(bloodPress)
				.append(SampleCLPmsg.PulseRt).append(pulseRate)
				.append(SampleCLPmsg.Temp).append(temp)
				.append(SampleCLPmsg.Wght).append(weight)
				.append(SampleCLPmsg.comment).append(comment);
			return buffy.toString();
		}

		buffy = new StringBuffer(dateStr);
		buffy.append(" ").append(timeStr).append(" ")
			.append(SampleCLPmsg.MedicalRecord);
		return buffy.toString();
	}
	private String readContacts(DataInputStream din) throws IOException
	{
		if (Debug.DEBUG && Debug.LEVEL > 5)
			System.out.println("  readContacts");
		// patientId and contactId are primary kesy
		String p = din.readUTF();	// skip patientId
		if (Debug.DEBUG && Debug.LEVEL > 5)
			System.out.println("\tpatientId " + p);

		String contactId = din.readUTF();
		if (Debug.DEBUG && Debug.LEVEL > 5)
			System.out.println("\tcontactId " + contactId);
		contactIds.addElement(contactId);

		din.readBoolean();
		String relationship = din.readUTF();

		String contactName = (String) persons.get(contactId);

		return contactName + " " + relationship;
	}

	private String [] vectorToStrArray(Vector v)
	{
		int size = 0;
		if (v == null || (size = v.size()) == 0)
			return null;

		// strings is global for reuse: new ChoiceGroup must size exactly
		if (strings == null || strings.length != size)
			strings = new String[size];

		for (int i = 0; i < size; i++)
		{
			strings[i] = (String) v.elementAt(i);
		}
		return strings;
	}

	private String[] readRecords(String storeName, int storeType, RecordFilter rf)
	{
		FastRecordStore rms = null;
		if (Debug.DEBUG && Debug.LEVEL > 5)
			System.out.println("readRecords "  + storeName);

		if (aVector == null)
			aVector = new Vector();
		else
			aVector.removeAllElements();
		try
		{
			rms = FastRecordStore.openRecordStore(storeName, false);
			FastRecordEnumeration enum =
							rms.enumerateRecords(null, null, false);
	
			if (Debug.DEBUG && Debug.LEVEL > 5)
				System.out.println("\t\treadRecords got enum # recs " 
					+ rms.getNumRecords());

			if (!enum.hasNextElement())
			{
				if (Debug.DEBUG && Debug.LEVEL > 5)
					System.out.println("\t\tno NextElement");
				return null;
			}
			// System.out.println(" readRecords hasNextElement");
			while (enum.hasNextElement())
			{
			// System.out.println(" readRecords get next id");
				int id = enum.nextRecordId();
			// System.out.println(" \t\t " + id);
				int recSize = -1;
				try
				{
					recSize = rms.getRecordSize(id);
					checkDataSizeOpenStream(recSize);
					recSize = rms.getRecord(id, data, 0);
				// com.ibm.mobileservices.util.ProtocolDumper.printBytes(data,
					// recSize, System.out, 32);
				}
				catch (Exception e)
				{
					if (Debug.DEBUG)
						System.out.println(e.toString());
				}
				// if (Debug.DEBUG)
					// System.out.println(" getRecord " + recSize);
				String s = null;
				byte dirty = din.readByte();
				if ((dirty & ISync.ROW_DELETED) != 0)
					continue;

				switch (storeType)
				{
				  case VNSCHEDULE:
					s = readSchedule(din);
				    break;
				  case VNPERSON:
				    s = readPerson(din);
				    break;
				  case VNMEDICALRECORD:
				    s = readMedRecord(id, din);
					    break;
				  case VNCONTACT:
					// not used, findRecords instead
				  	s = readContacts(din);
				    break;

				  default:
					if (Debug.DEBUG)
						System.out.println("Unknown store type");
					return null;
				}

				// if (Debug.DEBUG && s != null)
					// System.out.println(s);

				if (s != null)
					aVector.addElement(s);
			}
			enum.destroy();
			return vectorToStrArray(aVector);
		}
		catch (Exception e)
		{
			// if (Debug.DEBUG && Debug.LEVEL > 5)
				// System.out.println(" Failed toopen '" + storeName + "' " +
				// e.toString());
			// e.printStackTrace();

			// currentForm.append(e.toString());
			return null;
		}
		finally
		{
			try
			{
				if (rms != null)
					rms.closeRecordStore();
			}
			catch (Exception e) { }
		}
	}
	private String[] findRecords(String storeName, int storeType)
	{
		FastRecordStore rms = null;
		// System.out.println("findRecords" + storeName);

		if (aVector == null)
			aVector = new Vector();
		else
			aVector.removeAllElements();

		try
		{
			rms = FastRecordStore.openRecordStore(storeName, false);
			//rms.startEnum();
			FastRecordEnumeration enum = 
								rms.enumerateRecords(null, null, false);

			if (!enum.hasNextElement())
				return null;

			while (enum.hasNextElement())
			{
				int id = enum.nextRecordId();
				int recSize = 0;

				try
				{
					recSize = rms.getRecordSize(id);
					//byte[] enumData = enum.nextRecord();
					checkDataSizeOpenStream(recSize);
					recSize = rms.getRecord(id, data, 0);
				}
				catch (Exception e2)
				{
					if (Debug.DEBUG)
						System.out.println(e2.toString());
				}
				// if (Debug.DEBUG)
					// System.out.println(" getRecord " + recSize);

				boolean match = false;

				// discard 1-byte row marker
				byte dirty = din.readByte();

				if ((dirty & ISync.ROW_DELETED) != 0)
					continue;

				switch (storeType)
				{
				  case VNPERSON:
				    // System.out.println(SampleCLPmsg.VNPERSON);
				    match = matchFirstFieldPatientId(din);
					if (match)
					{
					if (Debug.DEBUG && Debug.LEVEL > 5)
						System.out.println(" MATCHED\n");
						((InputStream) din).reset();
						// discard 1-byte row marker
						din.readByte();
						String s = readPerson(din);
						aVector.addElement(s);
					}
					else
					if (Debug.DEBUG && Debug.LEVEL > 5)
						System.out.println(" Person No MATCH");
				    break;
				  case VNMEDICALRECORD:
				    match = matchMedRecord(din);
					if (match)
					{
						((InputStream) din).reset();

						// discard 1-byte row marker
						din.readByte();
						String s = readMedRecord(id, din);
						aVector.addElement(s);
					}
				    break;
				  case VNCONTACT:
					// System.out.println(SampleCLPmsg.VNCONTAC);
				    match = matchFirstFieldPatientId(din);
					if (match)
					{
						((InputStream) din).reset();
						// discard 1-byte row marker
						din.readByte();
				// com.ibm.mobileservices.util.ProtocolDumper.printBytes(data,
					// recSize, System.out, 32);
						String s = readContacts(din);
						aVector.addElement(s);
					}
				    break;

				  default:
					if (Debug.DEBUG && Debug.LEVEL > 5)
						System.out.println("Unknown store type");
					return null;
				}
				// only one person matches
				if (match && storeType == VNPERSON)
					break;
			}
			enum.destroy();
			return vectorToStrArray(aVector);
		}
		catch (Exception e)
		{
			currentForm.append(e.toString());
			if (Debug.DEBUG)
				System.out.println("findRecord failed " + e.toString());
			if (Debug.DEBUG && Debug.LEVEL > 8)
				e.printStackTrace();
			return null;
		}
		finally
		{
			try
			{
				if (rms != null)
					rms.closeRecordStore();
			}
			catch (Exception e) { }
		}
	}
	
	private boolean matchFirstFieldPatientId(DataInputStream din) throws IOException
	{
		// prim key not null din.readBoolean();
		String thisId = din.readUTF();
		if (Debug.DEBUG)
			System.out.println("matchFirstFieldPatientId thisId " + thisId + " patientId " + patientId);
		if (thisId.equals(patientId))
			return true;
		return false;
	}
	
	private boolean matchMedRecord(DataInputStream din) throws IOException
	{
		// Col 1
		int thisRecId = din.readInt();

		// track max for possible AddRecord by user
		if (thisRecId > maxRecId)
			maxRecId = thisRecId;

		// in verbose mode we search on a single *record* id
		if (medVerbose)
		{

// This code should be dead

			if (Debug.DEBUG)
				System.out.println("thisRecId " + thisRecId
					+ " recordId " + recordId);
			if (thisRecId == recordId)
				return true;
// This code should be dead -- end
		}
		else
		{
			// Col 2: otherwise search all records on *patient* id
			din.readBoolean();
			din.readInt();	// date

			// Col 3
			din.readBoolean();
			din.readInt();	// time

			// Col 4: no null byte after 7/19/2002
			// din.readBoolean();
			String thisId = din.readUTF();
			// if (Debug.DEBUG)
				// System.out.println("  thisPatientId " + thisId + " patientId " + patientId);
			if (thisId.equals(patientId))
				return true;
		}
		return false;
	}

	/**
		Return the new form in currentForm: caller must save it
		Return the new choiceGroup as the return of the method
	 */
	private ChoiceGroup showChoiceGroup(
		String[] choices, Form form, String formTitle,
		String choiceTitle, Command[] cmds)
			throws Exception
	{
		if (Debug.DEBUG)
			System.out.println("showChoiceGroup " + formTitle + " " + choiceTitle);

		if (form == null)
			form = new Form(formTitle);
		else
			clearDisplay(form);
		currentForm = form;

		ChoiceGroup choice = null;
		if (choices == null)
			currentForm.append("Nothing found in database");
		else
		{
			try
			{
				choice = new ChoiceGroup(
					choiceTitle, Choice.EXCLUSIVE, choices, null);
			}
			catch (Exception e)
			{
				if (Debug.DEBUG)
					System.out.println(" choiceGroup failed " + e.toString());
				if (Debug.DEBUG && Debug.LEVEL > 8)
					e.printStackTrace();
				throw e;
			}

			form.append(choice);
		}

		for (int i = 0; i < cmds.length; i++)
			form.addCommand(cmds[i]);

		display.setCurrent(form);

		return choice;
	}

	private Form showForm(
		String[] elements, Form form, String formTitle,
			String choiceTitle, Command[] cmds)
			throws Exception
	{
		if (Debug.DEBUG)
			System.out.println("showForm");

		if (form == null)
			form = new Form(formTitle);
		else
			clearDisplay(form);

		// form.setCommandListener(caller);

		if (Debug.DEBUG)
			System.out.println(" elements " + elements);
		for (int i = 0; elements != null && i < elements.length; i++)
			form.append(elements[i]);

		// form.setItemStateListener(caller);

		for (int i = 0; i < cmds.length; i++)
			form.addCommand(cmds[i]);

		display.setCurrent(form);

		return form;
	}

	private String fmtInt(int val)
	{
	// System.out.println("fmtInt val " + val);
		if (val < 10)
			return "0" + val;
		else 
			return "" + val;
	}

	// private String fmtTime(java.util.Date tm)
	public String fmtTime(int tm)
	{
		int hour = -1, min = -1;

		hour = tm >> 16;
		min = (tm >> 8) & 0xFF;
		// int sec = tm & 0xff;

		String hrStr = fmtInt(hour);
		String minStr = fmtInt(min);
		return hrStr + ":" + minStr;
	}
	// private String fmtDate(java.util.Date dt)
	public String fmtDate(int dt)
	{
		int day = -1, mon = -1, year = -1;

		year = dt >> 16;
		mon = (dt >> 8) & 0xFF;
		day =  dt & 0xFF;

		return new String (mon + "-" + day + "-" + year);
	}

	private void checkDataSizeOpenStream(int recSize)
		throws IOException
	{
		if (recSize <= data.length)
		{
			// open streams if required 
			if (bin == null)
				openStreams(data);

			// reset to start of data
			((InputStream) din).reset();
		    return;
		}

		data = new byte[recSize + 40];	// plan ahead, add 40

		// open on new data byte array
		if (din != null)
			((InputStream) din).close();

		openStreams(data);
	}
	private void openStreams(byte data[])
	{
		bin = new ByteArrayInputStream(data);
		din = new DataInputStream(bin);
	}

	private void clearDisplay()
	{
		clearDisplay(currentForm);
	}
	private void clearDisplay(Form f)
	{
		for (int i = f.size() - 1; i >= 0; i--)
			f.delete(i);
	}
	void checkStateChanged(ChoiceGroup cg, int index)
	{
		if (cg == schedChoice)
		{
			schedChoosen = index;
			if (Debug.DEBUG)
			{
				System.out.println("schedChoice: schedChoosen is "
					+ schedChoosen);
				if (schedChoosen < schedule.length)
					System.out.println("choice is " + schedule[schedChoosen]);
			}
		}
		else if (cg == recChoice)
		{
			recChoosen = index;
			Integer rmsRecId = (Integer) recordIds.elementAt(index);
			recordId = rmsRecId.intValue();
			if (Debug.DEBUG)
				System.out.println(">> recChoice: recChoosen " + recChoosen + " rec " + recordId);
		}
		else if (cg == contactChoice)
		{
			contactChoosen = index;
			if (Debug.DEBUG)
				System.out.println("contactChoice: contactChoosen is " + contactChoosen);
		}
	}
}
